/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef __DBMS_H__
#define __DBMS_H__

/* FIXME: Should we remove this? */
#include <mod_dav.h>
#include "lock.h"

#define dav_repos_COLLECTION 1
#define dav_repos_TEXT 1
#define dav_repos_BIN  2

struct dav_repos_property
{
    long serialno;
    int version;         /* For DeltaV 'version_property' table */
    long ns_id;
    const char *name;
    const char *value;
    struct dav_repos_property *next;
};

typedef struct dav_repos_property dav_repos_property;
typedef struct dav_repos_stream dav_repos_stream;

/*
** DBMS data transfer interface
*/
struct dav_repos_resource
{
    dav_resource_type type;

    long serialno;
    char *uri;
    long long creationdate;
    const char *displayname;
    const char *getcontentlanguage;
    const char *getcontenttype;
    long getcontentlength;
    const char *getetag;
    long long getlastmodified;  /* Extra. Not in database table */
    const char *lockdiscovery;
    int resourcetype;       /* 0:regular file; 1:collection(directory) */
    int source;         /* 0:regular; 1:source */
    const char *supportedlock;  /* Extra. Not in database table */
    int depth;
    int istext;         /* 0:none; 1: Text; 2: Binary */
    int isexternal;     /* 0:no, !0: yes */
    char *owner;        /* owner/principal property */

    apr_hash_t *lpr_hash;   /* live property hash */

    dav_repos_property *pr; /* for dead property data */

    apr_hash_t *pr_hash;    /* use for output value */
    apr_hash_index_t *pr_hash_index;    /* pr_hash_index to get first and next */

    apr_hash_t *ns_hash;    /* namespace(URI) to ns_id hash */
    apr_hash_t *ns_id_hash; /* ns_id to namespace(URI) hash */

    /* Version handling */
    apr_hash_t *vpr_hash;   /* version property hash */
    const char *root_version;   /* Root version props */
    const char *creator_displayname;
    const char *comment;

    int isversioned;            /* For DeltaV 'dasl_resource' table */
    int checkin;                /* For DeltaV 'dasl_resource' table */
    int checkout;               /* For DeltaV 'dasl_resource' table */
    int version;                /* For DeltaV 'version_resource' table*/
    int lastversion;            /* Checkin -1 or checkout value or root_version*/

    struct dav_repos_resource *next;    /* for multiful result */

    apr_pool_t *p;      /* apache pool */
    
};

typedef struct dav_repos_resource dav_repos_resource;

/* dav_repos_server_conf is in repos.h */
/* typedef dav_repos_server_conf dav_repos_db; */

/* context needed to identify a resource */
struct dav_resource_private
{
    apr_pool_t *pool;       /* memory storage pool associated with request */
    const char *pathname;   /* full pathname to resource */
    dav_repos_db *db;       /* DB related data */
    const dav_repos_resource *db_r;
    apr_finfo_t finfo;      /* filesystem info */
    request_rec *rec;           /* Save rec */
};

typedef struct dbms_stream dbms_stream;

/* define the dav_stream structure for our use */
struct dav_stream
{
    apr_pool_t *p;
    dbms_stream *stream;
    dav_repos_db *db;
    dav_repos_resource *db_r;
};


/* MYSQL functions */
int dbms_opendb ( dav_repos_db * d );
int dbms_closedb ( dav_repos_db * d );
int dbms_get_property ( dav_repos_db * d, dav_repos_resource * r );
long dbms_set_property ( dav_repos_db * d, const dav_repos_resource * r );
int dbms_write_content ( dav_repos_db * d,
                         dav_repos_resource * db_r, const char *tpfilename );
int dbms_read_content ( dav_repos_db * d,
                        dav_repos_resource * db_r, const char *tpfilename );

int dbms_fill_dead_property ( dav_repos_db * d,
                              dav_repos_resource * db_r );
int dbms_get_collection_resource ( dav_repos_db * d,
                                   dav_repos_resource * db_r, int depth );

/* exists check using pathname */
int dbms_exists ( dav_repos_db *db, const char *pathname,
                  apr_pool_t * pool );

/* I'll fill all d and r_src and r_des->uri */
int dbms_move_resource ( dav_repos_db * d,
                         const dav_repos_resource * r_src,
                         const dav_repos_resource * r_des );

/* I'll fill all d and r_src and r_des->uri */
int dbms_copy_resource ( dav_repos_db * d,
                         const dav_repos_resource * r_src,
                         const dav_repos_resource * r_des,
                         int depth );

/* Copys the dead properties of resources */
int dbms_copy_dead_properties ( dav_repos_db * d,
                                const dav_repos_resource * r_src,
                                const dav_repos_resource * r_des
                              );

/* I'll fill pr->name and you will fill pr->value */
int dbms_get_dead_property ( dav_repos_db * d,
                             const dav_repos_resource * r,
                             dav_repos_property * pr );

/* I'll fill pr->name and you will fill pr->value */
int dbms_get_all_dead_property ( dav_repos_db * d,
                                 const dav_repos_resource * r,
                                 dav_repos_property ** pr );

int db_insert_property ( dav_repos_db * d,
                         const dav_repos_resource * r,
                         const dav_repos_property * pr );

/* I'll fill pr->name and pr->value */
int dbms_set_dead_property ( dav_repos_db * d,
                             const dav_repos_resource * r,
                             const dav_repos_property * pr );


/* I'll fill pr->name */
int dbms_del_dead_property ( dav_repos_db * d,
                             const dav_repos_resource * r,
                             const dav_repos_property * pr );

/* remove all files if the resource is collection */
int dbms_remove_resource ( dav_repos_db * d,
                           const dav_repos_resource * r );
/* remove all files if the resource is collection */

const char *dav_repos_getetag_dbr ( const dav_repos_resource * db_r );
int dbms_get_resourcetype ( dav_repos_db * d, dav_repos_resource * r );
int dbms_search ( dav_repos_db * d, char *search_command,
                  dav_repos_resource * db_r );
/* Build namespace
 *
 * Read namespace infor from DB and fill hash
 * db_r must have pool
 * This hash is for PROPFIND
 * */
void dbms_build_ns_id_hash ( dav_repos_db * d, dav_repos_resource * db_r );

const char *dbms_get_ns ( dav_repos_db * d,
                          dav_repos_resource * db_r, const long ns_id );

long dbms_get_ns_id ( dav_repos_db * d,
                      dav_repos_resource * db_r, const char *namespace );


/* NULL Lock */

dav_error *dbms_load_locknull_list ( dav_repos_db * db,
                                     char *path,
                                     apr_text_header * thead,
                                     apr_pool_t * pool );

dav_error *dbms_save_locknull_member ( dav_repos_db * db,
                                       const char *filename, apr_pool_t * pool );

dav_error *dbms_remove_locknull_member ( dav_repos_db * db,
        const char *filename,
        apr_pool_t * pool );

dav_error *dbms_load_lock_record ( dav_repos_db * db,
                                   char *key,
                                   dav_lock_discovery ** direct,
                                   dav_lock_indirect ** indirect,
                                   int *need_save,
                                   apr_pool_t * pool );

dav_error *dbms_save_lock_record ( dav_repos_db * db,
                                   char *key,
                                   const dav_lock_discovery * direct,
                                   const dav_lock_indirect * indirect,
                                   apr_pool_t * pool );

dav_error *dbms_remove_lock_record ( dav_repos_db * db,
                                     char *key, apr_pool_t * pool );


int dbms_lock_exists ( dav_repos_db * db,
                       char *key, apr_pool_t * pool );

dav_error * dbms_set_is_versioned ( dav_repos_db * db, dav_repos_resource * db_r,
                                    int checkin );
dav_error * dbms_set_checkin_out ( dav_repos_db * db, dav_repos_resource * db_r,
                                   int checkin, int checkout );
dav_error * dbms_create_vr ( dav_repos_db * db, dav_repos_resource * db_r, int version );
dav_error * dbms_get_vrs ( dav_repos_db * db, dav_repos_resource * db_r,
                           dav_repos_resource ** vrs );

dav_error *dbms_open_stream ( dav_repos_db *db, dav_repos_resource *db_r,
                              dbms_stream **stream, dav_stream_mode mode );

dav_error *dbms_close_stream ( dbms_stream *stream, int commit );

dav_error *dbms_write_stream ( dbms_stream *stream, const void *buf,
                               const apr_size_t bufsize );

dav_error *dbms_seek_stream ( dbms_stream *stream, const apr_size_t pos );

dav_error *dbms_deliver ( dav_repos_db *db, dav_repos_resource *db_r,
                          ap_filter_t *output );

long dbms_insert_id (dav_repos_db *db, const char *tablename, const char *column);

ap_dbd_t *dbms_get_db(dav_repos_db *db);
#endif
