/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

/* TODO
1. remove etags filed
2. change collection to int
3. change all time field to long
*/

#include <httpd.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_core.h>      /* for ap_construct_url */
#include <mod_dav.h>

#include <apr_strings.h>
#include <apr_hash.h>
#include <apr_dbm.h>
#include <apr_file_io.h>

#include "dav_repos.h"
#include "dbms.h"
#include "util.h"


struct dav_namespace_map
{
    int *ns_map;

};


struct dav_db
{
    apr_pool_t *pool;
    dav_repos_db *mdb;
    dav_repos_resource *mdb_r;

    /* when used as a property database: */
    int version;        /* *minor* version of this db */

    dav_buffer ns_table;    /* table of namespace URIs */
    short ns_count;     /* number of entries in table */
    int ns_table_dirty;     /* ns_table was modified */
    apr_hash_t *uri_index;

    dav_buffer wb_key;      /* work buffer for dav_gdbm_key */

    apr_datum_t iter;       /* iteration key */
};

static dav_error *dav_repos_db_open ( apr_pool_t * p,
                                      const dav_resource * resource, int ro,
                                      dav_db ** pdb )
{
    dav_db *db = apr_pcalloc ( p, sizeof ( *db ) );

    TRACE();
    /* Set some values for DB */
    db->pool = p;
    db->mdb = resource->info->db;
    db->mdb_r = ( dav_repos_resource * ) resource->info->db_r;
    *pdb = db;

    db->uri_index = apr_hash_make ( p );

    return NULL;
}

static void dav_repos_db_close ( dav_db * db )
{
    TRACE();
}

static dav_error *dav_repos_db_define_namespaces ( dav_db * db,
        dav_xmlns_info * xi )
{
    int i;
    const char *ns, *xmlns;
    dav_repos_property *dead_prop;
    dav_repos_resource *dbr = db->mdb_r;
    char *ns_use;

    TRACE();


    /* Null lock */
    if ( dbr == NULL )
        return NULL;

    /* namespace flag */
    ns_use = apr_pcalloc ( dbr->p, DAV_REPOS_MAX_NAMESPACE );

    /* Check all used name space in this resource */
    for ( dead_prop = dbr->pr; dead_prop; dead_prop = dead_prop->next )
    {
        ns_use[ ( int ) dead_prop->ns_id] = 1;
    }


    /* Add namespace for dead properties */
    for ( i = 0; i < DAV_REPOS_MAX_NAMESPACE; i++ )
    {
        /* Is is used */
        if ( ns_use[i] == 1 )
        {
            ns = dbms_get_ns ( db->mdb, dbr, i );
            if ( ns && strlen ( ns ) )
            {   /* not nullns */
                xmlns = apr_psprintf ( dbr->p, "ns%d", i );
                dav_xmlns_add ( xi, xmlns, ns );
            }
        }
    }

    /* ### we don't have any other possible namespaces right now. */

    return NULL;
}

static dav_error *dav_repos_db_output_value ( dav_db * db,
        const dav_prop_name * name,
        dav_xmlns_info * xi,
        apr_text_header * phdr,
        int *found )
{
    const char *s = NULL;
    const char *val;

    /* Get resource records */
    dav_repos_resource *dbr = db->mdb_r;
    apr_pool_t *pool = db->pool;

    TRACE();

    if ( dbr == NULL )
    {
        *found = 0;
        return NULL;
    }

    DBG2 ( "OUT %s  [] %s\n", name->ns, name->name );

    if ( !strcmp ( name->ns, "DAV:" ) )
    {
        val = apr_hash_get ( dbr->lpr_hash, name->name, APR_HASH_KEY_STRING );
        if ( val )
            s = apr_psprintf ( pool, "<D:%s>%s</D:%s>" DEBUG_CR,
                               name->name, val, name->name );
        /* Maybe lockdiscovery */
        else if ( !strcmp ( name->name, "lockdiscovery" ) && dbr->lockdiscovery )
        {
            s = apr_psprintf ( pool, "<D:%s>%s</D:%s>" DEBUG_CR,
                               name->name, dbr->lockdiscovery, name->name );

        }
        else if ( !strcmp ( name->name, "supportedlock" ) && dbr->supportedlock )
        {
            s = apr_psprintf ( pool, "<D:%s>%s</D:%s>" DEBUG_CR,
                               name->name, dbr->supportedlock, name->name );

        }
        else if ( !strcmp ( name->name, "displayname" ) && dbr->supportedlock )
        {
            s = apr_psprintf ( pool, "<D:%s>%s</D:%s>" DEBUG_CR,
                               name->name, dbr->displayname, name->name );

        }
        /* Version props */
        if ( dbr->vpr_hash &&
                ( val = apr_hash_get ( dbr->vpr_hash, name->name, APR_HASH_KEY_STRING ) ) )
        {
            s = apr_psprintf ( pool, "<D:%s>%s</D:%s>" DEBUG_CR,
                               name->name, val, name->name );
        }
    }

    DBG2 ( "OUT2 %s  [] %s\n", name->ns, name->name );

    if ( s == NULL )
    {       /* Dead props or Not found */
        /* Get the key */
        const char *key =
            dav_repos_build_ns_name_key ( name->ns, name->name, pool );

        /* Get the dead prop */
        dav_repos_property *dead_prop =
            apr_hash_get ( dbr->pr_hash, key, APR_HASH_KEY_STRING );

        if ( dead_prop != NULL )
        {
            /* Null ns */
            if ( strlen ( name->ns ) == 0 || !strcmp ( name->ns, " " ) )
                s = apr_psprintf ( pool, "<%s xmlns=\"\">%s</%s>" DEBUG_CR,
                                   name->name, dead_prop->value, name->name );
            else        /* Not null ns */
                s = apr_psprintf ( pool, "<ns%ld:%s>%s</ns%ld:%s>" DEBUG_CR,
                                   dead_prop->ns_id, name->name,
                                   dead_prop->value,
                                   dead_prop->ns_id, name->name );
        }
        else
        {
            DBG1 ( "NOFPUND [%s]\n", key );
        }
    }

    if ( s )
    {
        *found = 1;
        apr_text_append ( pool, phdr, s );
    }
    else
    {
        *found = 0;
    }

    return NULL;
}

static dav_error *dav_repos_db_map_namespaces ( dav_db * db,
        const apr_array_header_t *
        namespaces,
        dav_namespace_map ** mapping )
{
    dav_namespace_map *m = apr_palloc ( db->pool, sizeof ( *m ) );
    int i;

    int *pmap;
    const char **puri;

    TRACE();

    /*
       ** Iterate over the provided namespaces. If a namespace already appears
       ** in our internal map of URI -> ns_id, then store that in the map. If
       ** we don't know the namespace yet, then add it to the map and to our
       ** table of known namespaces.
     */
    m->ns_map = pmap = apr_palloc ( db->pool, namespaces->nelts * sizeof ( *pmap ) );

    for ( i = namespaces->nelts, puri = ( const char ** ) namespaces->elts;
            i-- > 0; ++puri, ++pmap )
    {

        const char *uri = *puri;

        // Set right ns_id for uri;
        *pmap = dbms_get_ns_id ( db->mdb, db->mdb_r, uri );
        DBG1 ( "PMAP: %d", *pmap );
    }

    *mapping = m;

    return NULL;
}

static dav_error *dav_repos_db_store ( dav_db * db, const dav_prop_name * name,
                                       const apr_xml_elem * elem,
                                       dav_namespace_map * mapping )
{
    size_t l_val;
    char *buf;
    char *val;
    dav_repos_property *pr = apr_palloc ( db->pool, sizeof ( *pr ) );
    dav_repos_resource *dbr = db->mdb_r;

    TRACE();

    if ( dbr->type == DAV_RESOURCE_TYPE_VERSION ||
            dbr->type == DAV_RESOURCE_TYPE_HISTORY )
        return dav_new_error ( db->pool, HTTP_CONFLICT, 0,
                               "You can not set props for version resource." );


    /* Litmus test <aa xmlns=""> case */
    /*
       if(name->ns==NULL || strlen(name->ns)==0) {
       return dav_new_error(db->pool, HTTP_BAD_REQUEST, 0,
       "Invalid namespace declaration.");
       }
     */
    /* quote all the values in the element */
    /* ### be nice to do this without affecting the element itself */
    /* ### of course, the cast indicates Badness is occurring here */
    apr_xml_quote_elem ( db->pool, ( apr_xml_elem * ) elem );

    /* generate a text blob for the xml:lang plus the contents */
    apr_xml_to_text ( db->pool, elem, APR_XML_X2T_LANG_INNER, NULL,
                      mapping->ns_map, ( const char ** ) &val, &l_val );

    /* Warning : val[0] value is NULL, so you should use 'val+1' */

    /* Let's fill fill pr */
    pr->ns_id = dbms_get_ns_id ( db->mdb, db->mdb_r, name->ns );
    pr->name = name->name;
    pr->value = val + 1;

    if ( dbms_set_dead_property ( db->mdb, db->mdb_r, pr ) != 0 )
    {
        return dav_new_error ( db->pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while getting a dead property." );
    }

    buf = apr_psprintf ( db->pool, "I want to save [%ld:%s] [%s] {%d}",
                         pr->ns_id, pr->name, pr->value, mapping->ns_map[0] );

    DBG0 ( buf );
    return NULL;
}


static dav_error *dav_repos_db_remove ( dav_db * db, const dav_prop_name * name )
{
    char *buf;
    dav_repos_property *pr = apr_palloc ( db->pool, sizeof ( *pr ) );

    pr->ns_id = dbms_get_ns_id ( db->mdb, db->mdb_r, name->ns );
    pr->name = name->name;

    if ( dbms_del_dead_property ( db->mdb, db->mdb_r, pr ) != 0 )
    {
        return dav_new_error ( db->pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while deleting a dead property." );
    }

    TRACE();
    buf = apr_psprintf ( db->pool, "I want to delete [%s] ", pr->name );
    DBG0 ( buf );
    return NULL;
}


static int dav_repos_db_exists ( dav_db * db, const dav_prop_name * name )
{
    /* Get resource records */
    dav_repos_resource *dbr = db->mdb_r;

    /* Build the key */
    const char *key;

    TRACE();

    if ( dbr == NULL )
        return 0;

    key = dav_repos_build_ns_name_key ( name->ns, name->name, dbr->p );

    /* Get the the prop */
    if ( apr_hash_get ( dbr->pr_hash, key, APR_HASH_KEY_STRING ) )
        return 1;

    return 0;
}

static dav_error *dav_repos_db_first_name ( dav_db * db, dav_prop_name * pname )
{
    const char *key = NULL;
    apr_ssize_t klen;
    dav_repos_property *pr;

    /* Get resource records */
    dav_repos_resource *dbr = db->mdb_r;

    TRACE();

    if ( dbr == NULL )
    {
        pname->ns = NULL;
        pname->name = NULL;
        return NULL;
    }

    dbr->pr_hash_index = apr_hash_first ( db->pool, dbr->pr_hash );
    if ( dbr->pr_hash_index == NULL )
    {
        pname->ns = NULL;
        pname->name = NULL;
        return NULL;
    }

    apr_hash_this ( dbr->pr_hash_index, ( void * ) &key, &klen, ( void * ) &pr );

    /* Get ns and name from the key */
    dav_repos_get_ns_name_key ( key, pname, db->pool );

    return NULL;
}

static dav_error *dav_repos_db_next_name ( dav_db * db, dav_prop_name * pname )
{
    const char *key = NULL;
    apr_ssize_t klen;
    dav_repos_property *pr;

    /* Get resource records */
    dav_repos_resource *dbr = db->mdb_r;

    TRACE();
    pname->ns = NULL;
    pname->name = NULL;

    if ( dbr == NULL )
    {
        return NULL;
    }

    /* Run first_name first */
    if ( dbr->pr_hash_index == NULL )
        return NULL;

    dbr->pr_hash_index = apr_hash_next ( dbr->pr_hash_index );

    /* End of hash */
    if ( dbr->pr_hash_index == NULL )
        return NULL;

    apr_hash_this ( dbr->pr_hash_index, ( void * ) &key, &klen, ( void * ) &pr );

    /* Get ns and name from the key */
    dav_repos_get_ns_name_key ( key, pname, db->pool );

    return NULL;
}

static dav_error *dav_repos_db_get_rollback ( dav_db * db,
        const dav_prop_name * name,
        dav_deadprop_rollback **
        prollback )
{
    TRACE();
    return NULL;
}

static dav_error *dav_repos_db_apply_rollback ( dav_db * db,
        dav_deadprop_rollback *
        rollback )
{
    TRACE();
    return NULL;
}

const dav_hooks_propdb dav_repos_hooks_propdb =
    {
        dav_repos_db_open,
        dav_repos_db_close,
        dav_repos_db_define_namespaces,
        dav_repos_db_output_value,
        dav_repos_db_map_namespaces,
        dav_repos_db_store,
        dav_repos_db_remove,
        dav_repos_db_exists,
        dav_repos_db_first_name,
        dav_repos_db_next_name,
        dav_repos_db_get_rollback,
        dav_repos_db_apply_rollback,
    };
