/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_request.h>
#include <http_core.h>      /* for ap_construct_url */
#include <mod_dav.h>

#include <apr.h>
#include <apr_strings.h>
#include <apr_hash.h>
#include <apr_tables.h>
#include <apr_file_io.h>

#include "dav_repos.h"
#include "dbms.h"
#include "util.h"
#include "mod_dav_repos_version.h"


/* Check parent, for MKCOL, PUT */
static dav_error *dav_repos_get_parent_resource ( const dav_resource * resource,
        dav_resource **
        result_parent );

/* to assist in debugging mod_dav's GET handling */
const dav_hooks_liveprop dav_repos_hooks_liveprop;

/*
** The namespace URIs that we use. This list and the enumeration must
** stay in sync.
*/
static const char *const dav_repos_namespace_uris[] =
    {
        "DAV:",
        "http://apache.org/dav/props/",
        NULL            /* sentinel */
    };

enum
{
    dav_repos_URI_DAV,      /* the DAV: namespace URI */
    dav_repos_URI_MYPROPS   /* the namespace URI for our custom props */
};


/*
** The single property that we define (in the DAV_FS_URI_MYPROPS namespace)
*/
#define DAV_PROPID_FS_executable        1

static const dav_liveprop_spec dav_repos_props[] =
    {
        {
            dav_repos_URI_DAV,
            "creationdate",
            DAV_PROPID_creationdate,
            0
        },
        {
            dav_repos_URI_DAV,
            "displayname",
            DAV_PROPID_displayname,
            0},
        {
            dav_repos_URI_DAV,
            "getcontentlanguage",
            DAV_PROPID_getcontentlanguage,
            0},
        {
            dav_repos_URI_DAV,
            "getcontentlength",
            DAV_PROPID_getcontentlength,
            0},
        {
            dav_repos_URI_DAV,
            "getcontenttype",
            DAV_PROPID_getcontenttype,
            0},
        {
            dav_repos_URI_DAV,
            "getetag",
            DAV_PROPID_getetag,
            0},
        {
            dav_repos_URI_DAV,
            "getlastmodified",
            DAV_PROPID_getlastmodified,
            0},
        {
            dav_repos_URI_DAV,
            "resourcetype",
            DAV_PROPID_resourcetype,
            0},
        {
            dav_repos_URI_MYPROPS,
            "executable",
            DAV_PROPID_FS_executable,
            0               /* handled special in dav_fs_is_writable */
        },

        {0}             /* sentinel */
    };



const dav_liveprop_group dav_repos_liveprop_group =
    {
        dav_repos_props,
        dav_repos_namespace_uris,
        &dav_repos_hooks_liveprop
    };

/* Find live prop and return its propid */
int dav_repos_find_liveprop ( const dav_resource * resource,
                              const char *ns_uri, const char *name,
                              const dav_hooks_liveprop ** hooks )
{
    dav_repos_resource *db_r;

    /* don't try to find any liveprops if this isn't "our" resource */
    if ( resource->hooks != &dav_repos_hooks_repos )
        return 0;

    db_r = ( dav_repos_resource * ) resource->info->db_r;

    if ( db_r->lpr_hash == NULL )
    {
        DBG0 ( "Set LPR_HASH" );
        return 0;
    }


    TRACE();

    *hooks = dav_repos_liveprop_group.hooks;

    if ( !strcmp ( ns_uri, "DAV:" ) &&
            apr_hash_get ( db_r->lpr_hash, name, APR_HASH_KEY_STRING ) )
    {
        int i;
        /* find propid */
        for ( i = 0; dav_repos_props[i].name; i++ )
        {
            if ( !strcmp ( name, dav_repos_props[i].name ) )
                return dav_repos_props[i].propid;
        }
    }

    /* Not found */
    return 0;
}


/* Insert all live props from live prop hash */
void dav_repos_insert_all_liveprops ( request_rec * r,
                                      const dav_resource * resource,
                                      dav_prop_insert what,
                                      apr_text_header * phdr )
{
    int klen;
    const char *s;
    const char *key, *val;
    apr_hash_index_t *hindex;
    dav_repos_resource *db_r;

    /* don't try to find any liveprops if this isn't "our" resource */
    if ( resource->hooks != &dav_repos_hooks_repos )
        return;

    TRACE();

    db_r = ( dav_repos_resource * ) resource->info->db_r;

    if ( !resource->exists || db_r == NULL )
    {
        /* a lock-null resource */
        /*
           ** ### technically, we should insert empty properties. dunno offhand
           ** ### what part of the spec said this, but it was essentially thus:
           ** ### "the properties should be defined, but may have no value".
         */
        apr_text_append ( r->pool, phdr, "<D:resourcetype/>" );
        return;
    }

    /* Read live props from hash */
    for ( hindex = apr_hash_first ( r->pool, db_r->lpr_hash );
            hindex; hindex = apr_hash_next ( hindex ) )
    {
        apr_hash_this ( hindex, ( void * ) &key, &klen, ( void * ) &val );
        s = apr_psprintf ( r->pool, "<D:%s>%s</D:%s>" DEBUG_CR, key, val, key );
        apr_text_append ( r->pool, phdr, s );
    }

}

/*
** get resource and return to result_resource
*/
static dav_error *dav_repos_get_resource ( request_rec * r,
        const char *root_path,
        const char *label,
        int use_checked_in,
        dav_resource ** result_resource )
{
    /* Let's decide the type */
    char *s;
    int is_vhr;
    int vnum;
    dav_resource_private *ctx;
    dav_resource *resource;
    dav_repos_db *db;
    dav_repos_resource *db_r;
    apr_pool_t *pool = r->pool;
    TRACE();
    
    
    
    /* Create private resource context descriptor */
    ctx = apr_pcalloc ( pool, sizeof ( *ctx ) );

    /* Read data from server_config */

    db = dav_repos_get_db ( r );
    
    /* EXPERIMENTAL START */
    db->r = r;
	/* EXPERIMENTAL END */
    
    db_r = apr_pcalloc ( pool, sizeof ( *db_r ) );

    ctx->db = db;
    ctx->db_r = db_r;
    ctx->finfo = r->finfo;
    ctx->pool = pool;
    ctx->rec = r;

    /*
    ** If there is anything in the path_info, then this indicates that the
    ** entire path was not used to specify the file/dir. We want to append
    ** it onto the filename so that we get a "valid" pathname for null
    ** resources.
    */
    s = apr_pstrcat ( pool, r->filename, r->path_info, NULL );

    /* make sure the pathname does not have a trailing "/" */
    dav_repos_no_trail ( s );

    ctx->pathname = s;

    /* Create resource descriptor */
    resource = apr_pcalloc ( pool, sizeof ( *resource ) );

    /* let's fill db_r */
    db_r->p = pool;

    /* make sure the URI does not have a trailing "/" */
    s = apr_pstrdup ( pool, r->uri );
    dav_repos_no_trail ( s );

    db_r->uri = s;

    /* Check the URI */
    db_r->root_version = dav_repos_parse_vcr_uri ( pool, s, &vnum, &is_vhr );

    if ( db_r->root_version == NULL )
        db_r->type = resource->type = DAV_RESOURCE_TYPE_REGULAR;
    else if ( is_vhr )
        db_r->type = resource->type = DAV_RESOURCE_TYPE_HISTORY;
    else
        db_r->type = resource->type = DAV_RESOURCE_TYPE_VERSION;

    resource->info = ctx;
    resource->hooks = &dav_repos_hooks_repos;
    resource->pool = r->pool;

    
    
    /* exist check */
    if ( dbms_get_property ( db, db_r ) == 1 )
    {
        resource->exists = 1;

        /* Is collection ? */
        if ( db_r->resourcetype == dav_repos_COLLECTION )
        {
            resource->collection = APR_DIR;
            /* For fixup_dir */
            r->finfo.filetype = APR_DIR;
        }

        /* Do I need to set finfo ? */
        r->finfo.fname = s;
        r->finfo.size = db_r->getcontentlength;

        /* Is versioned?? */
        resource->versioned = db_r->isversioned;
        resource->working = ( db_r->checkout != -1 );
    }
    else
    {
        /* Calculate depth */
        db_r->depth = ap_count_dirs ( ( const char * ) s );
        resource->exists = 0;

        /* FIXME: Hum. Deal locknull */
        r->path_info = "";
    }

    /* make sure the URI does not have a trailing "/" */
    s = apr_pstrdup ( pool, r->uri );
    dav_repos_no_trail ( s );
    resource->uri = s;

    /* Set return value */
    *result_resource = resource;
    return NULL;
}

/* Should we return NULL, if no parent ? */
static dav_error *dav_repos_get_parent_resource ( const dav_resource * resource,
        dav_resource ** result_parent )
{
    dav_resource_private *ctx = resource->info;
    dav_resource_private *parent_ctx;
    dav_resource *parent_resource;
    dav_repos_db *db = resource->info->db;
    apr_pool_t *pool = resource->pool;
    dav_repos_resource *db_r;
    char *dirpath;
    char *uri;

    TRACE();

    /* EXPERIMENTAL START */
	db->r = resource->info->rec;
	/* EXPERIMENTAL END */
    
    /* If given resource is root, then there is no parent */
    if ( strcmp ( resource->uri, "/" ) == 0 || strcmp ( ctx->pathname, "/" ) == 0 )
    {
        *result_parent = NULL;
        return NULL;
    }

    /* ### optimize this into a single allocation! */

    /* Create private resource context descriptor */
    parent_ctx = apr_pcalloc ( pool, sizeof ( *parent_ctx ) );

    /* ### this should go away */
    parent_ctx->pool = pool;

    dirpath = ap_make_dirstr_parent ( pool, ctx->pathname );
    dav_repos_no_trail ( dirpath );
    parent_ctx->pathname = dirpath;

    /* Set parent resource */
    parent_resource = apr_pcalloc ( pool, sizeof ( *parent_resource ) );

    parent_resource->info = parent_ctx;
    parent_resource->hooks = &dav_repos_hooks_repos;
    parent_resource->pool = pool;

    if ( resource->uri == NULL )
        return NULL;

    /* Let's allow to delete version history */
    if ( resource->type == DAV_RESOURCE_TYPE_HISTORY )
        uri = apr_pstrdup ( pool, resource->uri );
    else
        uri = ap_make_dirstr_parent ( pool, resource->uri );

    dav_repos_no_trail ( uri );
    parent_resource->uri = uri;

    /* fill DBR to check collection type */
    db_r = apr_pcalloc ( resource->pool, sizeof ( *db_r ) );
    parent_ctx->db_r = db_r;

    db_r->p = pool;
    db_r->uri = apr_pstrdup ( resource->pool, parent_resource->uri );

    /* exist check */
    if ( dbms_get_property ( db, db_r ) == 1 )
    {
        parent_resource->exists = 1;

        /* Is collection ? */
        if ( db_r->resourcetype == dav_repos_COLLECTION )
        {
            parent_resource->collection = APR_DIR;
        }
    }
    else
    {
        parent_resource->exists = 0;
    }

    *result_parent = parent_resource;
    return NULL;
}

static int dav_repos_is_same_resource ( const dav_resource * res1,
                                        const dav_resource * res2 )
{
    TRACE();

    if ( res1->hooks != res2->hooks )
        return 0;

    return strcmp ( res1->uri, res2->uri ) == 0;
}

/* Check
**   parent : res1
**   child  : res2
*/
static int dav_repos_is_parent_resource ( const dav_resource * res1,
        const dav_resource * res2 )
{
    const char *parent_uri;
    TRACE();

    if ( res1->hooks != res2->hooks )
        return 0;

    if ( res1->exists == 0 )
        return 0;

    parent_uri = ap_make_dirstr_parent ( res2->pool, res2->uri );
    return strcmp ( res1->uri, parent_uri ) == 0;
}

/**
 * Note that this is only called for write (PUT).
 */
static dav_error *dav_repos_open_stream ( const dav_resource * resource,
        dav_stream_mode mode,
        dav_stream ** stream )
{
    dav_error *err;
    const char *slength;
    dav_resource *parent;
    request_rec *subrec;
    request_rec *rec = resource->info->rec;
    apr_pool_t *pool = resource->info->pool;
    dav_stream *ds = apr_pcalloc ( pool, sizeof ( *ds ) );
    dav_repos_db *db = resource->info->db;
    dav_repos_resource *db_r = ( dav_repos_resource * ) resource->info->db_r;

    /* EXPERIMENTAL START */
	db->r = resource->info->rec;
	/* EXPERIMENTAL END */
    
    TRACE();

    switch ( mode )
    {
        case DAV_MODE_WRITE_TRUNC:
        case DAV_MODE_WRITE_SEEKABLE:
            DBG0 ( "WRITE MODE" );

            /* Version resource */
            if ( resource->type ==DAV_RESOURCE_TYPE_VERSION ||
                    resource->type ==DAV_RESOURCE_TYPE_HISTORY )
                return dav_new_error ( pool, HTTP_CONFLICT, 0,
                                       "You can't update version resource or version history." );

            /* Check parent */
            /* FIXME : Should we check here? It resutns 403 insted of 409 */
            err = dav_repos_get_parent_resource ( resource, &parent );
            if ( err )
            {
                err = dav_push_error ( pool, err->status, 0,
                                       "Could not fetch parent resource information.",
                                       err );
                return err;
            }

            if ( parent->exists == 0 || parent->collection == 0 )
            {
                return dav_new_error ( pool, HTTP_CONFLICT, 0,
                                       "No parent collection." );
            }

            /* make time */
            db_r->getlastmodified = apr_time_now();

            /* set owner as username from Browser */
            db_r->owner = apr_pstrdup ( pool, rec->user );

            if ( db_r->creationdate == 0 )
                db_r->creationdate = db_r->getlastmodified;

            /* get length */
            slength = apr_table_get ( rec->headers_in, "Content-Length" );
            db_r->getcontentlength = slength ? atol ( slength ) : 0;

            // TODO: check to see if this works.
            /* First , try to get the content-type from client */

            if ( rec->content_type )
            {
                db_r->getcontenttype = apr_pstrdup ( pool, rec->content_type );
            }
            else
            {
                /* perform a "GET" on the resource's URI (note that the resource
                   may not correspond to the current request!). */
                subrec = ap_sub_req_lookup_uri ( rec->uri, rec, NULL );

                // useless: rec->uri does not exist yet.
                if ( subrec && subrec->content_type )
                {
                    db_r->getcontenttype = apr_pstrdup ( pool, subrec->content_type );
                }
            }

            /* No content type */
            if ( db_r->getcontenttype == NULL )
                db_r->getcontenttype = apr_pstrdup ( pool, "application/octet-stream" );

            DBG1 ( "CONTENT-TYPE: %s", db_r->getcontenttype );

            break;
    }

    ds->db = db;
    ds->db_r = db_r;
    if ( ( err = dbms_open_stream ( db, db_r, & ( ds->stream ), mode ) ) ) return err;

    *stream = ds;
    return NULL;
}


static dav_error *dav_repos_close_stream ( dav_stream * stream, int commit )
{
    TRACE();

    if ( commit )
    {
        /* Let's save db info here */
        if ( ( stream->db_r->serialno
                = dbms_set_property ( stream->db, stream->db_r ) ) == -1 )
        {
            dbms_close_stream ( stream->stream, 0 );

            /* Error handling */
            return dav_new_error ( stream->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                                   "database error" );
        }
    }

#ifdef LETS_USE_DBMS_CLOSE
    DBG1 ( "Deleting (Close) file : %s\n", filename );
    /* remove temprary file */
    if ( apr_file_remove ( filename, stream->p ) != APR_SUCCESS )
    {
        return dav_new_error ( stream->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "There was a problem removing (rolling "
                               "back) the resource "
                               "when it was being closed." );

    }
    return NULL;
#else
    return dbms_close_stream ( stream->stream, commit );
#endif
}

static dav_error *dav_repos_write_stream ( dav_stream * stream, const void *buf,
        apr_size_t bufsize )
{
    TRACE();

    return dbms_write_stream ( stream->stream, buf, bufsize );
}

static dav_error *dav_repos_seek_stream ( dav_stream * stream,
        apr_off_t abs_pos )
{
    TRACE();

    return dbms_seek_stream ( stream->stream, abs_pos );
}

static dav_error *dav_repos_set_headers ( request_rec * r,
        const dav_resource * resource )
{
    dav_repos_resource *db_r = ( dav_repos_resource * ) resource->info->db_r;

    TRACE();

    if ( !resource->exists )
        return NULL;

    /* make sure the proper mtime is in the request record */
    ap_update_mtime ( r, db_r->getlastmodified );

    /* ### note that these use r->filename rather than <resource> */
    ap_set_last_modified ( r );

    /* generate our etag and place it into the output */
    apr_table_setn ( r->headers_out, "ETag", dav_repos_getetag ( resource ) );

    /* we accept byte-ranges */
    apr_table_setn ( r->headers_out, "Accept-Ranges", "bytes" );

    /* For a directory, we will send text/html or text/xml. */
    if ( db_r->resourcetype == dav_repos_COLLECTION )
    {
        r->content_type = "text/html";
        r->no_local_copy = 1;
    }
    else if ( db_r->getcontenttype )
    {
        r->content_type = db_r->getcontenttype;
        r->no_local_copy = 1;
    }

    return NULL;
}


static dav_error *dav_repos_deliver ( const dav_resource * resource,
                                      ap_filter_t * output )
{
    char sep = '/';
    char *ztmp;

    apr_pool_t *pool = resource->pool;
    apr_bucket_brigade *bb;
    apr_status_t status;
    apr_bucket *bkt;

    dav_repos_db *db = resource->info->db;
    dav_repos_resource *tmp_r;
    dav_repos_resource *db_r = ( dav_repos_resource * ) resource->info->db_r;

    TRACE();

    /* Check resource type */
    if ( resource->type != DAV_RESOURCE_TYPE_REGULAR
            && resource->type != DAV_RESOURCE_TYPE_VERSION
            && resource->type != DAV_RESOURCE_TYPE_WORKING )
    {
        return dav_new_error ( pool, HTTP_CONFLICT, 0,
                               "Cannot GET this type of resource." );
    }

    if ( resource->collection )
    {
        bb = apr_brigade_create ( resource->pool, output->c->bucket_alloc );

        /* Will be filled children */
        if ( dbms_get_collection_resource ( db, db_r, 1 ) < 0 )
        {
            return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                                   "An error occurred while reading DBMS." );
        }

        ap_fprintf ( output, bb, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">" );
        ap_fprintf ( output, bb, "<html><head><title>Index of %s</title></head>\n"
                     "<body>\n <h2>Index of %s</h2>\n",
                     db_r->uri, db_r->uri );


        /* Add Up One Level */
        ztmp = apr_pstrndup ( pool, db_r->uri, strrchr ( db_r->uri, sep ) +1 - db_r->uri );
        ap_fprintf ( output, bb, "<a href=\"%s\"><small>Up One Level</small></a>", ztmp );
        ap_fprintf ( output, bb, "\n<ul>\n" );

        /*
        ** Lets walk through the results, 
        ** assemble walk resource, and call walker
        */
        for ( tmp_r = db_r; tmp_r; tmp_r = tmp_r->next )
        {
            /* assemble walk resource */

            /* Collection */
            if ( tmp_r->resourcetype == dav_repos_COLLECTION )
            {
                ap_fprintf ( output, bb, "<li><A HREF=\"%s/\">%s/</A><BR>\n",
                             tmp_r->uri,
                             tmp_r->uri );
            }
            else
            {
                ap_fprintf ( output, bb, "<li><A HREF=\"%s\">%s</A><BR>\n",
                             tmp_r->uri,
                             tmp_r->uri );
            }
        }

        ap_fputs ( output, bb,
                   " </ul>\n <hr noshade><em>Powered by "
                   "<a href=\"http://catacomb.tigris.org/\">Catacomb</a> "
                   "version " VERSION "."
                   "</em>\n</body></html>" );

        bkt = apr_bucket_eos_create ( output->c->bucket_alloc );
        APR_BRIGADE_INSERT_TAIL ( bb, bkt );
        if ( ( status = ap_pass_brigade ( output, bb ) ) != APR_SUCCESS )
        {
            return dav_new_error ( resource->pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                                   "Could not write EOS to filter." );
        }

        return NULL;

#if 0
        return dav_new_error ( pool, HTTP_CONFLICT, 0,
                               "There is no default response to GET for a "
                               "collection." );
#endif
    }

#ifdef NO_USE_DBMS_DELIVER
    /* Anyway we need to make random filename */
    filename = apr_psprintf ( pool, "%s/dav_repos_deliver_XXXXXX", db->tmp_dir );
    if ( mktemp ( filename ) == NULL )
    {
        return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while opening a resource name." );
    }

    /* get file from data base */
    if ( dbms_read_content ( db, db_r, filename ) != 0 )
    {
        return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while getting a resource name." );
    }

    /* open temparory file */
    if ( ( status = apr_file_open ( &fd, filename,
                                    APR_READ | APR_BINARY, 0,
                                    pool ) ) != APR_SUCCESS )
    {
        return dav_new_error ( pool, HTTP_FORBIDDEN, 0,
                               "File permissions deny server access." );
    }

    bb = apr_brigade_create ( pool, output->c->bucket_alloc );

    /* ### this does not handle large files. but this is test code anyway */
    bkt = apr_bucket_file_create ( fd, 0,
                                   ( apr_size_t ) db_r->getcontentlength,
                                   pool, output->c->bucket_alloc );
    APR_BRIGADE_INSERT_TAIL ( bb, bkt );

    bkt = apr_bucket_eos_create ( output->c->bucket_alloc );
    APR_BRIGADE_INSERT_TAIL ( bb, bkt );

    if ( ( status = ap_pass_brigade ( output, bb ) ) != APR_SUCCESS )
    {
        return dav_new_error ( pool, HTTP_FORBIDDEN, 0,
                               "Could not write contents to filter." );
    }

    /* ## FIXME: Should we delete file here? */
    DBG1 ( "Deleting (Filter) file : %s\n", filename );
    /* remove temprary file */
    if ( apr_file_remove ( filename, pool ) != APR_SUCCESS )
    {
        return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "There was a problem removing (rolling "
                               "back) the resource "
                               "when it was being closed." );

    }
    return NULL;
#else
    return dbms_deliver ( db, db_r, output );
#endif
}

static dav_error *dav_repos_create_collection ( dav_resource * resource )
{

    dav_repos_resource *db_r = ( dav_repos_resource * ) resource->info->db_r;
    dav_repos_db *db = resource->info->db;
    dav_resource *parent;
    dav_error *err;

    /* EXPERIMENTAL START */
	db->r = resource->info->rec;
	/* EXPERIMENTAL END */
    
    TRACE();

    if ( db_r->type == DAV_RESOURCE_TYPE_VERSION ||
            db_r->type == DAV_RESOURCE_TYPE_HISTORY )
        return dav_new_error ( db_r->p, HTTP_CONFLICT, 0,
                               "Version resource." );

    err = dav_repos_get_parent_resource ( resource, &parent );
    if ( err )
    {
        err = dav_push_error ( db_r->p, err->status, 0,
                               "Could not fetch parent resource information.",
                               err );
        return err;
    }

    if ( parent->exists == 0 || parent->collection == 0 )
    {
        return dav_new_error ( db_r->p, HTTP_CONFLICT, 0,
                               "No parent collection." );
    }

    /* let's fill db_r  for collection */
    db_r->getcontenttype = apr_pstrdup ( db_r->p, DIR_MAGIC_TYPE );
    db_r->resourcetype = dav_repos_COLLECTION;

    /* make time */
    db_r->creationdate = db_r->getlastmodified = apr_time_now();

    if ( ( db_r->serialno = dbms_set_property ( db, db_r ) ) == -1 )
    {
        return dav_new_error ( db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while writing to DBMS." );
    }

    return NULL;
}

static dav_error *dav_repos_copy_resource ( const dav_resource * src,
        dav_resource * dst,
        int depth, dav_response ** response )
{
    dav_error *err;
    dav_resource *parent;
    dav_repos_db *db = src->info->db;
    dav_repos_resource *db_r_src = ( dav_repos_resource * ) src->info->db_r;
    dav_repos_resource *db_r_dst = ( dav_repos_resource * ) dst->info->db_r;

    TRACE();

    /* If it's a version resource */
    if ( src->type == DAV_RESOURCE_TYPE_VERSION ||
            src->type == DAV_RESOURCE_TYPE_HISTORY ||
            dst->type == DAV_RESOURCE_TYPE_VERSION ||
            dst->type == DAV_RESOURCE_TYPE_HISTORY )
    {
        return dav_new_error ( db_r_src->p, HTTP_CONFLICT, 0,
                               "Can not copy or move version resource." );
    }

    /* Check dest's parent */
    err = dav_repos_get_parent_resource ( dst, &parent );
    if ( err )
    {
        err = dav_push_error ( db_r_src->p, err->status, 0,
                               "Could not fetch parent resource information.",
                               err );
        return err;
    }

    if ( parent->exists == 0 || parent->collection == 0 )
    {
        return dav_new_error ( db_r_src->p, HTTP_CONFLICT, 0,
                               "No parent collection." );
    }


    /* check src and des url */
    if ( strcmp ( db_r_src->uri, db_r_dst->uri ) == 0 )
    {
        return dav_new_error ( db_r_src->p, HTTP_BAD_REQUEST, 0,
                               "Source and destination are same while copying DBMS." );
    }

    /* copy resources */
    if ( dbms_copy_resource ( db, db_r_src, db_r_dst, depth ) != 0 )
    {
        return dav_new_error ( db_r_src->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while copying DBMS." );
    }
    return NULL;
}

static dav_error *dav_repos_move_resource ( dav_resource * src,
        dav_resource * dst,
        dav_response ** response )
{
    dav_error *err;
    dav_resource *parent;
    dav_repos_db *db = src->info->db;
    dav_repos_resource *db_r_src = ( dav_repos_resource * ) src->info->db_r;
    dav_repos_resource *db_r_dst = ( dav_repos_resource * ) dst->info->db_r;

    TRACE();

    /* If it's a version resource */
    if ( src->type == DAV_RESOURCE_TYPE_VERSION ||
            src->type == DAV_RESOURCE_TYPE_HISTORY ||
            dst->type == DAV_RESOURCE_TYPE_VERSION ||
            dst->type == DAV_RESOURCE_TYPE_HISTORY )
    {
        return dav_new_error ( db_r_src->p, HTTP_CONFLICT, 0,
                               "Can not copy or move version resource." );
    }


    /* Check dest's parent */
    err = dav_repos_get_parent_resource ( dst, &parent );
    if ( err )
    {
        err = dav_push_error ( db_r_src->p, err->status, 0,
                               "Could not fetch parent resource information.",
                               err );
        return err;
    }

    if ( parent->exists == 0 || parent->collection == 0 )
    {
        return dav_new_error ( db_r_src->p, HTTP_CONFLICT, 0,
                               "No parent collection." );
    }


    /* check src and des url */
    if ( strcmp ( db_r_src->uri, db_r_dst->uri ) == 0 )
    {
        return dav_new_error ( db_r_src->p, HTTP_BAD_REQUEST, 0,
                               "Source and destination are same while moveing DBMS." );
    }

    /* TODO: need a better implementation /test/move -> /test/movedest would fail */
    /*if ( strlen ( db_r_src->uri ) < strlen ( db_r_dst->uri ) )
    {
        char* tmp_str = NULL;
        tmp_str = apr_pstrmemdup ( db_r_src->p, db_r_dst->uri, strlen ( db_r_src->uri ) );
        if ( strcmp ( db_r_src->uri, tmp_str ) == 0 )
            return dav_new_error ( db_r_src->p, HTTP_BAD_REQUEST, 0,
                                   "Could not move to same destination - DMBS." );
    }*/

    /* move resource(s) */
    if ( dbms_move_resource ( db, db_r_src, db_r_dst ) != 0 )
    {
        return dav_new_error ( db_r_src->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while moving DBMS." );
    }

    return NULL;
}


static dav_error *dav_repos_remove_resource ( dav_resource * resource,
        dav_response ** response )
{
    dav_repos_db *db = resource->info->db;
    dav_repos_resource *db_r = ( dav_repos_resource * ) resource->info->db_r;

    TRACE();

    /* Version resource */
    if ( resource->type==DAV_RESOURCE_TYPE_VERSION )
    {
        return dav_new_error ( db_r->p, HTTP_CONFLICT, 0,
                               "It's Version resource" );
    }

    /* Delete the resource */
    if ( dbms_remove_resource ( db, db_r ) != 0 )
    {
        return dav_new_error ( db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "An error occurred while deleteing DBMS." );
    }

    return NULL;
}

static dav_error *dav_repos_walk ( const dav_walk_params * params, int depth,
                                   dav_response ** response )
{
    dav_error *err;
    apr_pool_t *pool = params->pool;
    dav_repos_db *db = params->root->info->db;
    dav_repos_resource *tmp_r;
    dav_repos_resource *db_r =
        ( dav_repos_resource * ) params->root->info->db_r;
    dav_walker_ctx *ctx = params->walk_ctx;

    /* EXPERIMENTAL START */
	db->r = params->root->info->rec;
	/* EXPERIMENTAL END */

    TRACE();

    DBG2 ( "Let's walk : %s in depth:%d", db_r->uri, depth );

    /* Let's start with NULL response */
    *response = NULL;

    /*
       ** search using mysql 
       ** if is not collection or depth=0, we have enough information
       ** We support only depth 0 and 1
     */
    if ( db_r->resourcetype == dav_repos_COLLECTION && depth != 0 )
    {
        /* Will be filled children */
        if ( dbms_get_collection_resource ( db, db_r, depth ) < 0 )
        {
            return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                                   "An error occurred while reading DBMS." );
        }
    }

    /*# FIXME : Only PROPFIND need this */
    /* Get properties in a form of property link for every resource */
    /* Should be one of those
       #define DAV_PROPFIND_IS_ALLPROP          1
       #define DAV_PROPFIND_IS_PROPNAME 2
       #define DAV_PROPFIND_IS_PROP             3
     */
    if ( ctx->propfind_type )
    {
        dbms_fill_dead_property ( db, db_r );

        /* Build name space id hash
        ** It will build db_r->ns_id_hash
        */
        dbms_build_ns_id_hash ( db, db_r );
    }

    /*
       ** Lets walk through the results, 
       ** assemble walk resource, and call walker
     */
    for ( tmp_r = db_r; tmp_r; tmp_r = tmp_r->next )
    {
        /* assemble walk resource */
        dav_walk_resource *wres = apr_pcalloc ( pool, sizeof ( *wres ) );
        dav_resource *resource = apr_pcalloc ( pool, sizeof ( *resource ) );
        dav_resource_private *info = apr_pcalloc ( pool, sizeof ( *info ) );

        /* Make info */
        info->pool = pool;
        info->pathname = tmp_r->uri;
        info->db = db;
        info->db_r = tmp_r;

        /* Make resource */
        resource->exists = params->root->exists;    /* nulllock */
        resource->collection = tmp_r->resourcetype;

        /* Add slash, if it is collection */
        /* Do we really need it? */
        /* We definitely don't, it causes a locking bug */
        /*if(tmp_r->resourcetype == dav_repos_COLLECTION)
            resource->uri = apr_pstrcat(pool, tmp_r->uri, "/", NULL);
        else */
        resource->uri = tmp_r->uri;

        resource->info = info;
        resource->hooks = params->root->hooks;
        resource->pool = pool;

        /* Make walk resource */
        wres->pool = pool;
        wres->resource = resource;
        wres->response = *response;
        wres->walk_ctx = params->walk_ctx;

        /* need to set pool */
        tmp_r->p = pool;

        /* Need to set ns_id  */
        tmp_r->ns_id_hash = db_r->ns_id_hash;

        /*
         * Build dead/live props hash
         * Build version props hash
         * It should be run even it's nulllock
         */
        dav_repos_build_pr_hash ( db, tmp_r );
        dav_repos_build_lpr_hash ( tmp_r );
        dav_repos_build_vpr_hash ( tmp_r );

        /* Fill lock discovery for propfind prop */
        if ( ctx->propfind_type == DAV_PROPFIND_IS_PROPNAME ||
                ctx->propfind_type == DAV_PROPFIND_IS_PROP )
            dav_repos_insert_lock_prop ( params, tmp_r );

        /* Call walker */
        if ( ( err = ( *params->func ) ( wres,
                                         tmp_r->resourcetype ==
                                         dav_repos_COLLECTION ?
                                         DAV_CALLTYPE_COLLECTION :
                                         DAV_CALLTYPE_MEMBER ) ) != NULL )
        {
            /* ### maybe add a higher-level description? */
            return err;
        }

        /* Save response for now */
        *response = wres->response;
    }

    /* Walk locknull files */
    if ( params->walk_type & DAV_WALKTYPE_LOCKNULL && depth )
    {

        apr_text_header phdr = { 0 };
        apr_text *t;

        DBG0 ( "LOCKNULL" );

        /* Let's read all locknull lists */
        err = dbms_load_locknull_list ( db, db_r->uri, &phdr, pool );
        if ( err != NULL )
            return err;

        /* For all lists */
        for ( t = phdr.first; t; t = t->next )
        {
            dav_lock *locks = NULL;
            /* assemble walk resource */
            dav_walk_resource *wres = apr_pcalloc ( pool, sizeof ( *wres ) );
            dav_resource *resource = apr_pcalloc ( pool, sizeof ( *resource ) );
            dav_resource_private *info = apr_pcalloc ( pool, sizeof ( *info ) );

#if 0
            dav_repos_build_pr_hash ( db, null_dbr );
            dav_repos_build_lpr_hash ( null_dbr );

            /* Fill lock discovery for propfind prop */
            if ( ctx->propfind_type == DAV_PROPFIND_IS_PROPNAME ||
                    ctx->propfind_type == DAV_PROPFIND_IS_PROP )
                dav_repos_insert_lock_prop ( params, null_dbr );

#endif
            /* Make info */
            info->pool = pool;
            info->pathname = t->text;
            info->db = db;
            /* Let's make null dbr */
            info->db_r = NULL;

            /* Make resource */
            resource->exists = 0;
            resource->uri = t->text;
            resource->info = info;
            resource->hooks = params->root->hooks;
            resource->pool = pool;

            /* Make walk resource */
            wres->pool = pool;
            wres->resource = resource;
            wres->response = *response;
            wres->walk_ctx = params->walk_ctx;

            if ( ( err = dav_lock_query ( params->lockdb, resource,
                                          &locks ) ) != NULL )
            {
                /* ### maybe add a higher-level description? */
                return err;
            }

            /* call the function for the specified dir + file */
            if ( locks != NULL &&
                    ( err = ( *params->func ) ( wres,
                                                DAV_CALLTYPE_LOCKNULL ) ) != NULL )
            {
                /* ### maybe add a higher-level description? */
                return err;
            }
            /* Save response for now */
            *response = wres->response;
        }

    }

    return NULL;
}

/*
**  make e tags using index and date
*/
const char *dav_repos_getetag_dbr ( const dav_repos_resource * db_r )
{
    return apr_psprintf ( db_r->p, "\"%lx-%lx-%lx\"",
                          ( unsigned long ) db_r->serialno,
                          ( unsigned long ) db_r->getcontentlength,
                          ( unsigned long ) db_r->creationdate );
}

const char *dav_repos_getetag ( const dav_resource * resource )
{
    TRACE();

    if ( !resource->exists || resource->info == NULL
            || resource->info->db_r == NULL )
        return apr_pstrdup ( resource->info->pool, "" );

    return dav_repos_getetag_dbr ( resource->info->db_r );
}

const dav_hooks_repository dav_repos_hooks_repos =
    {
        1,              /* special GET handling *//* 1 for GET handling, 0 for generic */
        dav_repos_get_resource,
        dav_repos_get_parent_resource,
        dav_repos_is_same_resource,
        dav_repos_is_parent_resource,
        dav_repos_open_stream,
        dav_repos_close_stream,
        dav_repos_write_stream,
        dav_repos_seek_stream,
        dav_repos_set_headers,
        dav_repos_deliver,
        dav_repos_create_collection,
        dav_repos_copy_resource,
        dav_repos_move_resource,
        dav_repos_remove_resource,
        dav_repos_walk,
        dav_repos_getetag,
    };

/* Isert prop with propid */
static dav_prop_insert dav_repos_insert_prop ( const dav_resource * resource,
        int propid, dav_prop_insert what,
        apr_text_header * phdr )
{
    int i;
    const char *value = NULL;
    const char *name = NULL;
    const char *s = NULL;
    apr_pool_t *pool = resource->pool;
    dav_repos_resource *dbr = ( dav_repos_resource * ) resource->info->db_r;

    TRACE();
    DBG1 ( "propid: %d", propid );

    /*
       ** None of FS provider properties are defined if the resource does not
       ** exist. Just bail for this case.
       **
       ** Even though we state that the FS properties are not defined, the
       ** client cannot store dead values -- we deny that thru the is_writable
       ** hook function.
     */
    if ( !resource->exists )
        return DAV_PROP_INSERT_NOTDEF;

    /* find propname using prop id */
    for ( i = 0; dav_repos_props[i].name; i++ )
    {
        if ( propid == dav_repos_props[i].propid )
        {
            name = dav_repos_props[i].name;
            break;
        }
    }

    /* ### what the heck was this property? */
    if ( name == NULL )
        return DAV_PROP_INSERT_NOTDEF;

    /* Get value */
    value = apr_hash_get ( dbr->lpr_hash, name, APR_HASH_KEY_STRING );

    /* ### Not found in the hash */
    if ( value == NULL )
        return DAV_PROP_INSERT_NOTDEF;

    /* Do something according to what */
    if ( what == DAV_PROP_INSERT_VALUE )
    {
        s = apr_psprintf ( pool, "<D:%s>%s</D:%s>" DEBUG_CR, name, value, name );
    }
    else if ( what == DAV_PROP_INSERT_NAME )
    {
        s = apr_psprintf ( pool, "<D:%s/>" DEBUG_CR, name );
    }
    else
    {
        /* assert: what == DAV_PROP_INSERT_SUPPORTED */
        s = apr_psprintf ( pool,
                           "<D:supported-live-property D:name=\"%s\" "
                           "D:namespace=\"%s\"/>" DEBUG_CR,
                           name, dav_repos_namespace_uris[0] );
    }
    apr_text_append ( pool, phdr, s );

    /* we inserted what was asked for */
    return what;
}

static int dav_repos_is_writable ( const dav_resource * resource, int propid )
{
    int i;

    TRACE();

    /* Try to find is_writable */
    for ( i = 0; dav_repos_props[i].name; i++ )
    {
        if ( propid == dav_repos_props[i].propid )
        {
            return dav_repos_props[i].is_writable;
        }
    }

    /* It is writable */
    return 1;
}

static dav_error *dav_repos_patch_validate ( const dav_resource * resource,
        const apr_xml_elem * elem,
        int operation,
        void **context, int *defer_to_dead )
{
    TRACE();
    return NULL;
}


static dav_error *dav_repos_patch_exec ( const dav_resource * resource,
        const apr_xml_elem * elem,
        int operation,
        void *context,
        dav_liveprop_rollback ** rollback_ctx )
{
    TRACE();
    return NULL;
}

static void dav_repos_patch_commit ( const dav_resource * resource,
                                     int operation,
                                     void *context,
                                     dav_liveprop_rollback * rollback_ctx )
{
    TRACE();
    /* nothing to do */
}


static dav_error *dav_repos_patch_rollback ( const dav_resource * resource,
        int operation,
        void *context,
        dav_liveprop_rollback *
        rollback_ctx )
{
    TRACE();
    return NULL;
}

const dav_hooks_liveprop dav_repos_hooks_liveprop =
    {
        dav_repos_insert_prop,
        dav_repos_is_writable,
        dav_repos_namespace_uris,
        dav_repos_patch_validate,
        dav_repos_patch_exec,
        dav_repos_patch_commit,
        dav_repos_patch_rollback
    };
